# -*- coding: utf-8 -*-


# python std lib
from __future__ import with_statement

import asyncio
import os

# 3rd party imports
import pytest
from mock import Mock, patch

# rediscluster imports
from coredis import StrictRedis
from coredis.connection import ClusterConnection, Connection, UnixDomainSocketConnection
from coredis.exceptions import RedisClusterException
from coredis.pool import ClusterConnectionPool, ConnectionPool

try:
    pass

    ssl_available = True
except ImportError:
    ssl_available = False


class DummyConnection:
    description_format = "DummyConnection<>"

    def __init__(self, host="localhost", port=7000, socket_timeout=None, **kwargs):
        self.kwargs = kwargs
        self.pid = os.getpid()
        self.host = host
        self.port = port
        self.socket_timeout = socket_timeout
        self.awaiting_response = False


class TestConnectionPool:
    async def get_pool(
        self,
        connection_kwargs=None,
        max_connections=None,
        max_connections_per_node=None,
        connection_class=DummyConnection,
    ):
        connection_kwargs = connection_kwargs or {}
        pool = ClusterConnectionPool(
            connection_class=connection_class,
            max_connections=max_connections,
            max_connections_per_node=max_connections_per_node,
            startup_nodes=[{"host": "127.0.0.1", "port": 7000}],
            **connection_kwargs
        )
        await pool.initialize()
        return pool

    @pytest.mark.asyncio()
    async def test_in_use_not_exists(self):
        """
        Test that if for some reason, the node that it tries to get the connectino for
        do not exists in the _in_use_connection variable.
        """
        pool = await self.get_pool()
        pool._in_use_connections = {}
        pool.get_connection("pubsub", channel="foobar")

    @pytest.mark.asyncio()
    async def test_connection_creation(self):
        connection_kwargs = {"foo": "bar", "biz": "baz"}
        pool = await self.get_pool(connection_kwargs=connection_kwargs)
        connection = pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})
        assert isinstance(connection, DummyConnection)
        for key in connection_kwargs:
            assert connection.kwargs[key] == connection_kwargs[key]

    @pytest.mark.asyncio()
    async def test_multiple_connections(self):
        pool = await self.get_pool()
        c1 = pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})
        c2 = pool.get_connection_by_node({"host": "127.0.0.1", "port": 7001})
        assert c1 != c2

    @pytest.mark.asyncio()
    async def test_max_connections(self):
        pool = await self.get_pool(max_connections=2)
        pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})
        pool.get_connection_by_node({"host": "127.0.0.1", "port": 7001})
        with pytest.raises(RedisClusterException):
            pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})

    @pytest.mark.asyncio()
    async def test_max_connections_per_node(self):
        pool = await self.get_pool(max_connections=2, max_connections_per_node=True)
        pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})
        pool.get_connection_by_node({"host": "127.0.0.1", "port": 7001})
        pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})
        pool.get_connection_by_node({"host": "127.0.0.1", "port": 7001})
        with pytest.raises(RedisClusterException):
            pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})

    @pytest.mark.asyncio()
    async def test_max_connections_default_setting(self):
        pool = await self.get_pool(max_connections=None)
        assert pool.max_connections == 2 ** 31

    @pytest.mark.asyncio()
    async def test_reuse_previously_released_connection(self):
        pool = await self.get_pool()
        c1 = pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})
        pool.release(c1)
        c2 = pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})
        assert c1 == c2

    @pytest.mark.asyncio()
    async def test_repr_contains_db_info_tcp(self):
        """
        Note: init_slot_cache muts be set to false otherwise it will try to
              query the test server for data and then it can't be predicted reliably
        """
        connection_kwargs = {"host": "127.0.0.1", "port": 7000}
        pool = await self.get_pool(
            connection_kwargs=connection_kwargs, connection_class=ClusterConnection
        )
        expected = "ClusterConnectionPool<ClusterConnection<host=127.0.0.1,port=7000>>"
        assert repr(pool) == expected

    @pytest.mark.asyncio()
    async def test_get_connection_by_key(self):
        """
        This test assumes that when hashing key 'foo' will be sent to server with port 7002
        """
        pool = await self.get_pool(connection_kwargs={})

        # Patch the call that is made inside the method to allow control of the returned
        # connection object
        with patch.object(
            ClusterConnectionPool, "get_connection_by_slot", autospec=True
        ) as pool_mock:

            def side_effect(self, *args, **kwargs):
                return DummyConnection(port=1337)

            pool_mock.side_effect = side_effect

            connection = pool.get_connection_by_key("foo")
            assert connection.port == 1337

        with pytest.raises(RedisClusterException) as ex:
            pool.get_connection_by_key(None)
        assert str(ex.value).startswith(
            "No way to dispatch this command to Redis Cluster."
        ), True

    @pytest.mark.asyncio()
    async def test_get_connection_by_slot(self):
        """
        This test assumes that when doing keyslot operation on "foo" it will return 12182
        """
        pool = await self.get_pool(connection_kwargs={})

        # Patch the call that is made inside the method to allow control of the returned
        # connection object
        with patch.object(
            ClusterConnectionPool, "get_connection_by_node", autospec=True
        ) as pool_mock:

            def side_effect(self, *args, **kwargs):
                return DummyConnection(port=1337)

            pool_mock.side_effect = side_effect

            connection = pool.get_connection_by_slot(12182)
            assert connection.port == 1337

        m = Mock()
        pool.get_random_connection = m

        # If None value is provided then a random node should be tried/returned
        pool.get_connection_by_slot(None)
        m.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_get_connection_blocked(self):
        """
        Currently get_connection() should only be used by pubsub command.
        All other commands should be blocked and exception raised.
        """
        pool = await self.get_pool()

        with pytest.raises(RedisClusterException) as ex:
            pool.get_connection("GET")
        assert str(ex.value).startswith(
            "Only 'pubsub' commands can use get_connection()"
        )

    @pytest.mark.asyncio()
    async def test_master_node_by_slot(self):
        pool = await self.get_pool(connection_kwargs={})
        node = pool.get_master_node_by_slot(0)
        node["port"] = 7000
        node = pool.get_master_node_by_slot(12182)
        node["port"] = 7002

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_connection_idle_check(self, event_loop):
        pool = ClusterConnectionPool(
            startup_nodes=[dict(host="127.0.0.1", port=7000)],
            max_idle_time=0.2,
            idle_check_interval=0.1,
        )
        conn = pool.get_connection_by_node(
            {
                "name": "127.0.0.1:7000",
                "host": "127.0.0.1",
                "port": 7000,
                "server_type": "master",
            }
        )
        name = conn.node["name"]
        assert len(pool._in_use_connections[name]) == 1
        # not ket could be found in dict for now
        assert not pool._available_connections
        pool.release(conn)
        assert len(pool._in_use_connections[name]) == 0
        assert len(pool._available_connections[name]) == 1
        await asyncio.sleep(0.21)
        assert len(pool._in_use_connections[name]) == 0
        assert len(pool._available_connections[name]) == 0
        last_active_at = conn.last_active_at
        assert last_active_at == conn.last_active_at
        assert conn._writer is None and conn._reader is None


class TestReadOnlyConnectionPool:
    async def get_pool(
        self, connection_kwargs=None, max_connections=None, startup_nodes=None
    ):
        startup_nodes = startup_nodes or [{"host": "127.0.0.1", "port": 7000}]
        connection_kwargs = connection_kwargs or {}
        pool = ClusterConnectionPool(
            max_connections=max_connections,
            startup_nodes=startup_nodes,
            readonly=True,
            **connection_kwargs
        )
        await pool.initialize()
        return pool

    @pytest.mark.asyncio()
    async def test_repr_contains_db_info_readonly(self):
        """
        Note: init_slot_cache must be set to false otherwise it will try to
              query the test server for data and then it can't be predicted reliably
        """
        pool = await self.get_pool(
            startup_nodes=[
                {"host": "127.0.0.1", "port": 7000},
                {"host": "127.0.0.2", "port": 7001},
            ],
        )
        assert "ClusterConnection<host=127.0.0.2,port=7001>" in repr(pool)
        assert "ClusterConnection<host=127.0.0.1,port=7000>" in repr(pool)

    @pytest.mark.asyncio()
    async def test_max_connections(self):
        pool = await self.get_pool(max_connections=2)
        pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})
        pool.get_connection_by_node({"host": "127.0.0.1", "port": 7001})
        with pytest.raises(RedisClusterException):
            pool.get_connection_by_node({"host": "127.0.0.1", "port": 7000})

    @pytest.mark.asyncio()
    async def test_get_node_by_slot(self):
        """
        We can randomly get all nodes in readonly mode.
        """
        pool = await self.get_pool(connection_kwargs={})

        expected_ports = set(range(7000, 7006))
        actual_ports = set()
        for i in range(0, 16383):
            node = pool.get_node_by_slot(i)
            actual_ports.add(node["port"])
        assert actual_ports == expected_ports


class TestConnectionPoolURLParsing:
    def test_defaults(self):
        pool = ConnectionPool.from_url("redis://localhost")
        assert pool.connection_class == Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 0,
            "password": None,
        }

    def test_hostname(self):
        pool = ConnectionPool.from_url("redis://myhost")
        assert pool.connection_class == Connection
        assert pool.connection_kwargs == {
            "host": "myhost",
            "port": 6379,
            "db": 0,
            "password": None,
        }

    def test_quoted_hostname(self):
        pool = ConnectionPool.from_url(
            "redis://my %2F host %2B%3D+", decode_components=True
        )
        assert pool.connection_class == Connection
        assert pool.connection_kwargs == {
            "host": "my / host +=+",
            "port": 6379,
            "db": 0,
            "password": None,
        }

    def test_port(self):
        pool = ConnectionPool.from_url("redis://localhost:6380")
        assert pool.connection_class == Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6380,
            "db": 0,
            "password": None,
        }

    def test_password(self):
        pool = ConnectionPool.from_url("redis://:mypassword@localhost")
        assert pool.connection_class == Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 0,
            "password": "mypassword",
        }

    def test_quoted_password(self):
        pool = ConnectionPool.from_url(
            "redis://:%2Fmypass%2F%2B word%3D%24+@localhost", decode_components=True
        )
        assert pool.connection_class == Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 0,
            "password": "/mypass/+ word=$+",
        }

    def test_quoted_path(self):
        pool = ConnectionPool.from_url(
            "unix://:mypassword@/my%2Fpath%2Fto%2F..%2F+_%2B%3D%24ocket",
            decode_components=True,
        )
        assert pool.connection_class == UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/my/path/to/../+_+=$ocket",
            "db": 0,
            "password": "mypassword",
        }

    def test_db_as_argument(self):
        pool = ConnectionPool.from_url("redis://localhost", db="1")
        assert pool.connection_class == Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 1,
            "password": None,
        }

    def test_db_in_path(self):
        pool = ConnectionPool.from_url("redis://localhost/2", db="1")
        assert pool.connection_class == Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 2,
            "password": None,
        }

    def test_db_in_querystring(self):
        pool = ConnectionPool.from_url("redis://localhost/2?db=3", db="1")
        assert pool.connection_class == Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 3,
            "password": None,
        }

    def test_extra_querystring_options(self):
        pool = ConnectionPool.from_url("redis://localhost?a=1&b=2")
        assert pool.connection_class == Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 0,
            "password": None,
            "a": "1",
            "b": "2",
        }

    def test_client_creates_connection_pool(self):
        r = StrictRedis.from_url("redis://myhost")
        assert r.connection_pool.connection_class == Connection
        assert r.connection_pool.connection_kwargs == {
            "host": "myhost",
            "port": 6379,
            "db": 0,
            "password": None,
        }


class TestConnectionPoolUnixSocketURLParsing:
    def test_defaults(self):
        pool = ConnectionPool.from_url("unix:///socket")
        assert pool.connection_class == UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 0,
            "password": None,
        }

    def test_password(self):
        pool = ConnectionPool.from_url("unix://:mypassword@/socket")
        assert pool.connection_class == UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 0,
            "password": "mypassword",
        }

    def test_db_as_argument(self):
        pool = ConnectionPool.from_url("unix:///socket", db=1)
        assert pool.connection_class == UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 1,
            "password": None,
        }

    def test_db_in_querystring(self):
        pool = ConnectionPool.from_url("unix:///socket?db=2", db=1)
        assert pool.connection_class == UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 2,
            "password": None,
        }

    def test_extra_querystring_options(self):
        pool = ConnectionPool.from_url("unix:///socket?a=1&b=2")
        assert pool.connection_class == UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 0,
            "password": None,
            "a": "1",
            "b": "2",
        }



# -*- coding: utf-8 -*-

# python std lib
import asyncio
import json
import os
import sys

# 3rd party imports
import pytest
from packaging import version

# rediscluster imports
from coredis import StrictRedis, StrictRedisCluster

# put our path in front so we can be sure we are testing locally not against the global package
basepath = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(1, basepath)

_REDIS_VERSIONS = {}


def get_versions(**kwargs):
    key = json.dumps(kwargs)

    if key not in _REDIS_VERSIONS:
        client = _get_client(**kwargs)
        loop = asyncio.get_event_loop()
        info = loop.run_until_complete(client.info())
        _REDIS_VERSIONS[key] = {
            key: value["redis_version"] for key, value in info.items()
        }

    return _REDIS_VERSIONS[key]


def skip_if_server_version_lt(min_version):
    """ """
    versions = get_versions()

    for v in versions.values():
        if version.parse(v) < version.parse(min_version):
            return pytest.mark.skipif(True, reason="")

    return pytest.mark.skipif(False, reason="")


def _get_client(cls=None, **kwargs):
    if not cls:
        cls = StrictRedisCluster

    params = {
        "startup_nodes": [{"host": "127.0.0.1", "port": 7000}],
        "stream_timeout": 10,
    }
    params.update(kwargs)

    return cls(**params)


def _init_mgt_client(request, cls=None, **kwargs):
    """ """
    client = _get_client(cls=cls, **kwargs)

    if request:

        def teardown():
            client.connection_pool.disconnect()

        request.addfinalizer(teardown)

    return client


def skip_if_not_password_protected_nodes():
    """ """

    return pytest.mark.skipif("TEST_PASSWORD_PROTECTED" not in os.environ, reason="")


@pytest.fixture()
def o(request, *args, **kwargs):
    """
    Create a StrictRedisCluster instance with decode_responses set to True.
    """
    params = {"decode_responses": True}
    params.update(kwargs)

    return _get_client(cls=StrictRedisCluster, **params)


@pytest.fixture()
async def r(request, *args, **kwargs):
    """
    Create a StrictRedisCluster instance with default settings.
    """

    client = _get_client(cls=StrictRedisCluster, **kwargs)
    await client.config_set("maxmemory-policy", "noeviction")
    await client.flushdb()

    return client


@pytest.fixture()
def ro(request, *args, **kwargs):
    """
    Create a StrictRedisCluster instance with readonly mode
    """
    params = {"readonly": True}
    params.update(kwargs)

    return _get_client(cls=StrictRedisCluster, **params)


@pytest.fixture()
async def s(*args, **kwargs):
    """
    Create a StrictRedisCluster instance with 'init_slot_cache' set to false
    """
    s = _get_client(**kwargs)
    assert s.connection_pool.nodes.slots == {}
    assert s.connection_pool.nodes.nodes == {}

    return s


@pytest.fixture()
async def t(*args, **kwargs):
    """
    Create a regular StrictRedis object instance
    """

    client = StrictRedis(*args, **kwargs)
    await client.flushdb()

    return client


@pytest.fixture()
async def sr(request, *args, **kwargs):
    """
    Returns a instance of StrictRedisCluster
    """

    client = _get_client(reinitialize_steps=1, cls=StrictRedisCluster, **kwargs)
    await client.flushdb()

    return client



#!/usr/bin/python
# -*- coding: utf-8 -*-

import asyncio
import os
import random
import threading
import time
import warnings
from itertools import chain
from urllib.parse import parse_qs, unquote, urlparse

from coredis.connection import (
    ClusterConnection,
    Connection,
    RedisSSLContext,
    UnixDomainSocketConnection,
)
from coredis.exceptions import ConnectionError, RedisClusterException
from coredis.nodemanager import NodeManager

FALSE_STRINGS = ("0", "F", "FALSE", "N", "NO")


def to_bool(value):
    if value is None or value == "":
        return None
    if isinstance(value, str) and value.upper() in FALSE_STRINGS:
        return False
    return bool(value)


URL_QUERY_ARGUMENT_PARSERS = {
    "stream_timeout": float,
    "connect_timeout": float,
    "retry_on_timeout": to_bool,
}


class ConnectionPool:
    """Generic connection pool"""

    @classmethod
    def from_url(cls, url, db=None, decode_components=False, **kwargs):
        """
        Returns a connection pool configured from the given URL.

        For example:

        - ``redis://[:password]@localhost:6379/0``
        - ``rediss://[:password]@localhost:6379/0``
        - ``unix://[:password]@/path/to/socket.sock?db=0``

        Three URL schemes are supported:

        - ``redis://`` <http://www.iana.org/assignments/uri-schemes/prov/redis>`_ creates a
          normal TCP socket connection
        - ``rediss://`` <http://www.iana.org/assignments/uri-schemes/prov/rediss>`_ creates a
          SSL wrapped TCP socket connection
        - ``unix://`` creates a Unix Domain Socket connection

        There are several ways to specify a database number. The parse function
        will return the first specified option:
        1. A ``db`` querystring option, e.g. redis://localhost?db=0
        2. If using the redis:// scheme, the path argument of the url, e.g.
        redis://localhost/0
        3. The ``db`` argument to this function.

        If none of these options are specified, db=0 is used.

        The ``decode_components`` argument allows this function to work with
        percent-encoded URLs. If this argument is set to ``True`` all ``%xx``
        escapes will be replaced by their single-character equivalents after
        the URL has been parsed. This only applies to the ``hostname``,
        ``path``, and ``password`` components.

        Any additional querystring arguments and keyword arguments will be
        passed along to the ConnectionPool class's initializer. The querystring
        arguments ``connect_timeout`` and ``stream_timeout`` if supplied
        are parsed as float values. The arguments ``retry_on_timeout`` are
        parsed to boolean values that accept True/False, Yes/No values to indicate state.
        Invalid types cause a ``UserWarning`` to be raised.
        In the case of conflicting arguments, querystring arguments always win.
        """
        url = urlparse(url)
        qs = url.query

        url_options = {}

        for name, value in iter(parse_qs(qs).items()):
            if value and len(value) > 0:
                parser = URL_QUERY_ARGUMENT_PARSERS.get(name)
                if parser:
                    try:
                        url_options[name] = parser(value[0])
                    except (TypeError, ValueError):
                        warnings.warn(
                            UserWarning(
                                "Invalid value for `%s` in connection URL." % name
                            )
                        )
                else:
                    url_options[name] = value[0]

        if decode_components:
            password = unquote(url.password) if url.password else None
            path = unquote(url.path) if url.path else None
            hostname = unquote(url.hostname) if url.hostname else None
        else:
            password = url.password
            path = url.path
            hostname = url.hostname

        # We only support redis:// and unix:// schemes.
        if url.scheme == "unix":
            url_options.update(
                {
                    "password": password,
                    "path": path,
                    "connection_class": UnixDomainSocketConnection,
                }
            )

        else:
            url_options.update(
                {"host": hostname, "port": int(url.port or 6379), "password": password}
            )

            # If there's a path argument, use it as the db argument if a
            # querystring value wasn't specified
            if "db" not in url_options and path:
                try:
                    url_options["db"] = int(path.replace("/", ""))
                except (AttributeError, ValueError):
                    pass

            if url.scheme == "rediss":
                keyfile = url_options.pop("ssl_keyfile", None)
                certfile = url_options.pop("ssl_certfile", None)
                cert_reqs = url_options.pop("ssl_cert_reqs", None)
                ca_certs = url_options.pop("ssl_ca_certs", None)
                url_options["ssl_context"] = RedisSSLContext(
                    keyfile, certfile, cert_reqs, ca_certs
                ).get()

        # last shot at the db value
        url_options["db"] = int(url_options.get("db", db or 0))

        # update the arguments from the URL values
        kwargs.update(url_options)
        return cls(**kwargs)

    def __init__(
        self,
        connection_class=Connection,
        max_connections=None,
        max_idle_time=0,
        idle_check_interval=1,
        **connection_kwargs
    ):
        """
        Creates a connection pool. If max_connections is set, then this
        object raises redis.ConnectionError when the pool's limit is reached.

        By default, TCP connections are created connection_class is specified.
        Use redis.UnixDomainSocketConnection for unix sockets.

        Any additional keyword arguments are passed to the constructor of
        connection_class.
        """
        max_connections = max_connections or 2 ** 31
        if not isinstance(max_connections, int) or max_connections < 0:
            raise ValueError('"max_connections" must be a positive integer')

        self.connection_class = connection_class
        self.connection_kwargs = connection_kwargs
        self.max_connections = max_connections
        self.max_idle_time = max_idle_time
        self.idle_check_interval = idle_check_interval
        self.loop = self.connection_kwargs.get("loop")

        self.reset()

    def __repr__(self):
        return "{}<{}>".format(
            type(self).__name__,
            self.connection_class.description.format(**self.connection_kwargs),
        )

    async def disconnect_on_idle_time_exceeded(self, connection):
        while True:
            if (
                time.time() - connection.last_active_at > self.max_idle_time
                and not connection.awaiting_response
            ):
                connection.disconnect()
                try:
                    self._available_connections.remove(connection)
                except ValueError:
                    pass
                self._created_connections -= 1
                break
            await asyncio.sleep(self.idle_check_interval)

    def reset(self):
        self.pid = os.getpid()
        self._created_connections = 0
        self._available_connections = []
        self._in_use_connections = set()
        self._check_lock = threading.Lock()

    def _checkpid(self):
        if self.pid != os.getpid():
            with self._check_lock:
                if self.pid == os.getpid():
                    # another thread already did the work while we waited
                    # on the lock.
                    return
                self.disconnect()
                self.reset()

    def get_connection(self, *args, **kwargs):
        """Gets a connection from the pool"""
        self._checkpid()
        try:
            connection = self._available_connections.pop()
        except IndexError:
            connection = self.make_connection()
        self._in_use_connections.add(connection)
        return connection

    def make_connection(self):
        """Creates a new connection"""
        if self._created_connections >= self.max_connections:
            raise ConnectionError("Too many connections")
        self._created_connections += 1
        connection = self.connection_class(**self.connection_kwargs)
        if self.max_idle_time > self.idle_check_interval > 0:
            # do not await the future
            asyncio.ensure_future(self.disconnect_on_idle_time_exceeded(connection))
        return connection

    def release(self, connection):
        """Releases the connection back to the pool"""
        self._checkpid()
        if connection.pid != self.pid:
            return
        self._in_use_connections.remove(connection)
        # discard connection with unread response
        if connection.awaiting_response:
            connection.disconnect()
            self._created_connections -= 1
        else:
            self._available_connections.append(connection)

    def disconnect(self):
        """Closes all connections in the pool"""
        all_conns = chain(self._available_connections, self._in_use_connections)
        for connection in all_conns:
            connection.disconnect()
            self._created_connections -= 1


class ClusterConnectionPool(ConnectionPool):
    """Custom connection pool for rediscluster"""

    RedisClusterDefaultTimeout = None

    def __init__(
        self,
        startup_nodes=None,
        connection_class=ClusterConnection,
        max_connections=None,
        max_connections_per_node=False,
        reinitialize_steps=None,
        skip_full_coverage_check=False,
        nodemanager_follow_cluster=False,
        readonly=False,
        max_idle_time=0,
        idle_check_interval=1,
        **connection_kwargs
    ):
        """
        :skip_full_coverage_check:
            Skips the check of cluster-require-full-coverage config, useful for clusters
            without the CONFIG command (like aws)
        :nodemanager_follow_cluster:
            The node manager will during initialization try the last set of nodes that
            it was operating on. This will allow the client to drift along side the cluster
            if the cluster nodes move around alot.
        """
        super(ClusterConnectionPool, self).__init__(
            connection_class=connection_class, max_connections=max_connections
        )

        # Special case to make from_url method compliant with cluster setting.
        # from_url method will send in the ip and port through a different variable then the
        # regular startup_nodes variable.
        if startup_nodes is None:
            if "port" in connection_kwargs and "host" in connection_kwargs:
                startup_nodes = [
                    {
                        "host": connection_kwargs.pop("host"),
                        "port": str(connection_kwargs.pop("port")),
                    }
                ]

        self.max_connections = max_connections or 2 ** 31
        self.max_connections_per_node = max_connections_per_node
        self.nodes = NodeManager(
            startup_nodes,
            reinitialize_steps=reinitialize_steps,
            skip_full_coverage_check=skip_full_coverage_check,
            max_connections=self.max_connections,
            nodemanager_follow_cluster=nodemanager_follow_cluster,
            **connection_kwargs
        )
        self.initialized = False

        self.connections = {}
        self.connection_kwargs = connection_kwargs
        self.connection_kwargs["readonly"] = readonly
        self.readonly = readonly
        self.max_idle_time = max_idle_time
        self.idle_check_interval = idle_check_interval
        self.reset()

        if "stream_timeout" not in self.connection_kwargs:
            self.connection_kwargs[
                "stream_timeout"
            ] = ClusterConnectionPool.RedisClusterDefaultTimeout

    def __repr__(self):
        """
        Returns a string with all unique ip:port combinations that this pool
        is connected to
        """
        return "{0}<{1}>".format(
            type(self).__name__,
            ", ".join(
                [
                    self.connection_class.description.format(**node)
                    for node in self.nodes.startup_nodes
                ]
            ),
        )

    async def initialize(self):
        if not self.initialized:
            await self.nodes.initialize()
            self.initialized = True

    async def disconnect_on_idle_time_exceeded(self, connection):
        while True:
            if (
                time.time() - connection.last_active_at > self.max_idle_time
                and not connection.awaiting_response
            ):
                connection.disconnect()
                node = connection.node
                self._available_connections[node["name"]].remove(connection)
                self._created_connections_per_node[node["name"]] -= 1
                break
            await asyncio.sleep(self.idle_check_interval)

    def reset(self):
        """Resets the connection pool back to a clean state"""
        self.pid = os.getpid()
        self._created_connections_per_node = {}  # Dict(Node, Int)
        self._available_connections = {}  # Dict(Node, List)
        self._in_use_connections = {}  # Dict(Node, Set)
        self._check_lock = threading.Lock()
        self.initialized = False

    def _checkpid(self):
        if self.pid != os.getpid():
            with self._check_lock:
                if self.pid == os.getpid():
                    # another thread already did the work while we waited
                    # on the lockself.
                    return
                self.disconnect()
                self.reset()

    def get_connection(self, command_name, *keys, **options):
        # Only pubsub command/connection should be allowed here
        if command_name != "pubsub":
            raise RedisClusterException(
                "Only 'pubsub' commands can use get_connection()"
            )

        channel = options.pop("channel", None)

        if not channel:
            return self.get_random_connection()

        slot = self.nodes.keyslot(channel)
        node = self.get_master_node_by_slot(slot)

        self._checkpid()

        try:
            connection = self._available_connections.get(node["name"], []).pop()
        except IndexError:
            connection = self.make_connection(node)

        if node["name"] not in self._in_use_connections:
            self._in_use_connections[node["name"]] = set()

        self._in_use_connections[node["name"]].add(connection)

        return connection

    def make_connection(self, node):
        """Creates a new connection"""
        if self.count_all_num_connections(node) >= self.max_connections:
            if self.max_connections_per_node:
                raise RedisClusterException(
                    "Too many connection ({0}) for node: {1}".format(
                        self.count_all_num_connections(node), node["name"]
                    )
                )

            raise RedisClusterException("Too many connections")

        self._created_connections_per_node.setdefault(node["name"], 0)
        self._created_connections_per_node[node["name"]] += 1
        connection = self.connection_class(
            host=node["host"], port=node["port"], **self.connection_kwargs
        )

        # Must store node in the connection to make it eaiser to track
        connection.node = node
        if self.max_idle_time > self.idle_check_interval > 0:
            # do not await the future
            asyncio.ensure_future(self.disconnect_on_idle_time_exceeded(connection))
        return connection

    def release(self, connection):
        """Releases the connection back to the pool"""
        self._checkpid()
        if connection.pid != self.pid:
            return

        # Remove the current connection from _in_use_connection and add it back to the available
        # pool. There is cases where the connection is to be removed but it will not exist and
        # there must be a safe way to remove
        i_c = self._in_use_connections.get(connection.node["name"], set())
        if connection in i_c:
            i_c.remove(connection)
        else:
            pass
        # discard connection with unread response
        if connection.awaiting_response:
            connection.disconnect()
            # reduce node connection count in case of too many connection error raised
            if (
                self.max_connections_per_node
                and self._created_connections_per_node.get(connection.node["name"])
            ):
                self._created_connections_per_node[connection.node["name"]] -= 1
        else:
            self._available_connections.setdefault(connection.node["name"], []).append(
                connection
            )

    def disconnect(self):
        """Closes all connectins in the pool"""
        all_conns = chain(
            self._available_connections.values(),
            self._in_use_connections.values(),
        )

        for node_connections in all_conns:
            for connection in node_connections:
                connection.disconnect()

    def count_all_num_connections(self, node):
        if self.max_connections_per_node:
            return self._created_connections_per_node.get(node["name"], 0)

        return sum([i for i in self._created_connections_per_node.values()])

    def get_random_connection(self):
        """Opens new connection to random redis server"""
        if self._available_connections:
            node_name = random.choice(list(self._available_connections.keys()))
            conn_list = self._available_connections[node_name]
            # check it in case of empty connection list
            if conn_list:
                return conn_list.pop()
        for node in self.nodes.random_startup_node_iter():
            connection = self.get_connection_by_node(node)

            if connection:
                return connection

        raise Exception("Cant reach a single startup node.")

    def get_connection_by_key(self, key):
        if not key:
            raise RedisClusterException(
                "No way to dispatch this command to Redis Cluster."
            )

        return self.get_connection_by_slot(self.nodes.keyslot(key))

    def get_connection_by_slot(self, slot):
        """
        Determines what server a specific slot belongs to and return a redis
        object that is connected
        """
        self._checkpid()

        try:
            return self.get_connection_by_node(self.get_node_by_slot(slot))
        except KeyError:
            return self.get_random_connection()

    def get_connection_by_node(self, node):
        """Gets a connection by node"""
        self._checkpid()
        self.nodes.set_node_name(node)

        try:
            # Try to get connection from existing pool
            connection = self._available_connections.get(node["name"], []).pop()
        except IndexError:
            connection = self.make_connection(node)

        self._in_use_connections.setdefault(node["name"], set()).add(connection)

        return connection

    def get_master_node_by_slot(self, slot):
        return self.nodes.slots[slot][0]

    def get_node_by_slot(self, slot):
        if self.readonly:
            return random.choice(self.nodes.slots[slot])
        return self.get_master_node_by_slot(slot)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import pytest

from coredis.exceptions import RedisError, ResponseError


class TestStreams:
    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xadd_with_wrong_id(self, r):
        with pytest.raises(RedisError):
            await r.xadd(
                "test_stream",
                {"k1": "v1", "k2": 1},
                stream_id=0,
                max_len=10,
                approximate=False,
            )

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xadd_without_given_id(self, r):
        stream_id = await r.xadd("test_stream", {"k1": "v1", "k2": 1})
        assert len(stream_id.decode().split("-")) == 2

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xadd_with_given_id(self, r):
        stream_id = await r.xadd(
            "test_stream", {"k1": "v1", "k2": 1}, stream_id="12321"
        )
        assert stream_id == b"12321-0"
        await r.flushdb()
        stream_id = await r.xadd(
            "test_stream", {"k1": "v1", "k2": 1}, stream_id="12321-0"
        )
        assert stream_id == b"12321-0"

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xadd_with_maxlen_accurately(self, r):
        for idx in range(10):
            await r.xadd(
                "test_stream", {"k1": "v1", "k2": 1}, max_len=2, approximate=False
            )
        # also test xlen here
        length = await r.xlen("test_stream")
        assert length == 2

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xadd_with_maxlen_approximately(self, r):
        for idx in range(10):
            await r.xadd(
                "test_stream", {"k1": "v1", "k2": 1}, max_len=2, approximate=True
            )
        length = await r.xlen("test_stream")
        assert length == 10

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xrange(self, r):
        for idx in range(1, 10):
            await r.xadd(
                "test_stream",
                {"k1": "v1", "k2": 1},
                stream_id=idx,
                max_len=10,
                approximate=False,
            )
        entries = await r.xrange("test_stream", count=5)
        assert (
            len(entries) == 5
            and isinstance(entries, list)
            and isinstance(entries[0], tuple)
        )
        entries = await r.xrange("test_stream", start="2", end="3", count=3)
        assert len(entries) == 2 and entries[0][0] == b"2-0"

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xrevrange(self, r):
        for idx in range(1, 10):
            await r.xadd(
                "test_stream",
                {"k1": "v1", "k2": 1},
                stream_id=idx,
                max_len=10,
                approximate=False,
            )
        entries = await r.xrevrange("test_stream", count=5)
        assert (
            len(entries) == 5
            and isinstance(entries, list)
            and isinstance(entries[0], tuple)
        )
        entries = await r.xrevrange("test_stream", start="2", end="3", count=3)
        assert len(entries) == 0
        entries = await r.xrevrange("test_stream", start="3", end="2", count=3)
        assert len(entries) == 2 and entries[0][0] == b"3-0"

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xread(self, r):
        for idx in range(1, 10):
            await r.xadd("test_stream", {"k1": "v1", "k2": 1}, stream_id=idx)
        entries = await r.xread(count=5, block=10, test_stream="0")
        assert len(entries[b"test_stream"]) == 5
        entries = await r.xread(count=10, block=10, test_stream="$")
        assert not entries
        entries = await r.xread(count=10, block=10, test_stream="2")
        assert entries and len(entries[b"test_stream"]) == 7
        assert entries[b"test_stream"][0] == (b"3-0", {b"k1": b"v1", b"k2": b"1"})

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xreadgroup(self, r):
        for idx in range(1, 10):
            await r.xadd("test_stream", {"k1": "v1", "k2": 1}, stream_id=idx)
        # read from group does not exist
        with pytest.raises(ResponseError):
            await r.xreadgroup(
                "wrong_group", "lalala", count=10, block=10, test_stream="1"
            )
        assert await r.xgroup_create("test_stream", "test_group", "0") is True
        entries = await r.xreadgroup(
            "test_group", "consumer1", count=5, test_stream=">"
        )
        assert len(entries[b"test_stream"]) == 5

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xgroup_create(self, r):
        for idx in range(1, 10):
            await r.xadd("test_stream", {"k1": "v1", "k2": 1}, stream_id=idx)
        with pytest.raises(ResponseError):
            await r.xgroup_create("wrong_group", "test_group")
        res = await r.xgroup_create("test_stream", "test_group")
        assert res is True
        group_info = await r.xinfo_groups("test_stream")
        assert len(group_info) == 1
        assert group_info[0][b"name"] == b"test_group"

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xgroup_set_id(self, r):
        for idx in range(1, 10):
            await r.xadd("test_stream", {"k1": "v1", "k2": 1}, stream_id=idx)
        assert await r.xgroup_create("test_stream", "test_group", "$") is True
        entries = await r.xreadgroup(
            "test_group", "consumer1", count=5, test_stream="1"
        )
        assert len(entries[b"test_stream"]) == 0
        group_info = await r.xinfo_groups("test_stream")
        assert group_info[0][b"pending"] == 0
        assert await r.xgroup_set_id("test_stream", "test_group", "0") is True
        await r.xreadgroup("test_group", "consumer1", count=5, test_stream=">")
        group_info = await r.xinfo_groups("test_stream")
        assert group_info[0][b"pending"] == 5

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xgroup_destroy(self, r):
        await r.xadd("test_stream", {"k1": "v1", "k2": 1})
        assert await r.xgroup_create("test_stream", "test_group") is True
        group_info = await r.xinfo_groups("test_stream")
        assert len(group_info) == 1
        assert group_info[0][b"name"] == b"test_group"
        assert await r.xgroup_destroy("test_stream", "test_group") == 1
        group_info = await r.xinfo_groups("test_stream")
        assert len(group_info) == 0

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xgroup_del_consumer(self, r):
        await r.xadd("test_stream", {"k1": "v1", "k2": 1})
        assert await r.xgroup_create("test_stream", "test_group") is True
        await r.xreadgroup("test_group", "consumer1", count=5, test_stream="1")
        group_info = await r.xinfo_groups("test_stream")
        assert len(group_info) == 1
        assert group_info[0][b"consumers"] == 1
        consumer_info = await r.xinfo_consumers("test_stream", "test_group")
        assert len(consumer_info) == 1
        assert consumer_info[0][b"name"] == b"consumer1"
        await r.xgroup_del_consumer("test_stream", "test_group", "consumer1")
        consumer_info = await r.xinfo_consumers("test_stream", "test_group")
        assert len(consumer_info) == 0

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_xpending(self, r):
        for idx in range(1, 10):
            await r.xadd("test_stream", {"k1": "v1", "k2": 1}, stream_id=idx)
        assert await r.xgroup_create("test_stream", "test_group", "$") is True
        entries = await r.xreadgroup(
            "test_group", "consumer1", count=5, test_stream="1"
        )
        assert len(entries[b"test_stream"]) == 0
        group_info = await r.xinfo_groups("test_stream")
        assert group_info[0][b"pending"] == 0
        assert (
            len(
                await r.xpending(
                    "test_stream", "test_group", count=10, consumer="consumer1"
                )
            )
            == 0
        )
        assert await r.xgroup_set_id("test_stream", "test_group", "0") is True
        await r.xreadgroup("test_group", "consumer1", count=5, test_stream=">")
        group_info = await r.xinfo_groups("test_stream")
        assert group_info[0][b"pending"] == 5
        assert (
            len(
                await r.xpending(
                    "test_stream", "test_group", count=10, consumer="consumer1"
                )
            )
            == 5
        )
        xpending_entries_in_range = await r.xpending(
            "test_stream",
            "test_group",
            start="2",
            end="2",
            count=10,
            consumer="consumer1",
        )
        assert len(xpending_entries_in_range) == 1
        assert xpending_entries_in_range[0][0] == b"2-0"



import datetime

from coredis.exceptions import RedisError
from coredis.utils import (
    NodeFlag,
    b,
    bool_ok,
    dict_merge,
    list_keys_to_dict,
    nativestr,
    pairs_to_dict,
    string_keys_to_dict,
)


def parse_slowlog_get(response, **options):
    return [
        {
            "id": item[0],
            "start_time": int(item[1]),
            "duration": int(item[2]),
            "command": b(" ").join(item[3]),
        }
        for item in response
    ]


def parse_client_list(response, **options):
    clients = []

    for c in nativestr(response).splitlines():
        # Values might contain '='
        clients.append(dict([pair.split("=", 1) for pair in c.split(" ")]))

    return clients


def parse_config_get(response, **options):
    response = [nativestr(i) if i is not None else None for i in response]

    return response and pairs_to_dict(response) or {}


def timestamp_to_datetime(response):
    """Converts a unix timestamp to a Python datetime object"""

    if not response:
        return None
    try:
        response = int(response)
    except ValueError:
        return None

    return datetime.datetime.fromtimestamp(response)


def parse_debug_object(response):
    """
    Parses the results of Redis's DEBUG OBJECT command into a Python dict
    """
    # The 'type' of the object is the first item in the response, but isn't
    # prefixed with a name
    response = nativestr(response)
    response = "type:" + response
    response = dict([kv.split(":") for kv in response.split()])

    # parse some expected int values from the string response
    # note: this cmd isn't spec'd so these may not appear in all redis versions
    int_fields = ("refcount", "serializedlength", "lru", "lru_seconds_idle")

    for field in int_fields:
        if field in response:
            response[field] = int(response[field])

    return response


def parse_info(response):
    """Parses the result of Redis's INFO command into a Python dict"""
    info = {}
    response = nativestr(response)

    def get_value(value):
        if "," not in value or "=" not in value:
            try:
                if "." in value:
                    return float(value)
                else:
                    return int(value)
            except ValueError:
                return value
        else:
            sub_dict = {}

            for item in value.split(","):
                k, v = item.rsplit("=", 1)
                sub_dict[k] = get_value(v)

            return sub_dict

    for line in response.splitlines():
        if line and not line.startswith("#"):
            if line.find(":") != -1:
                key, value = line.split(":", 1)
                info[key] = get_value(value)
            else:
                # if the line isn't splittable, append it to the "__raw__" key
                info.setdefault("__raw__", []).append(line)

    return info


def parse_role(response):
    role = nativestr(response[0])

    def _parse_master(response):
        offset, slaves = response[1:]
        res = {"role": role, "offset": offset, "slaves": []}

        for slave in slaves:
            host, port, offset = slave
            res["slaves"].append(
                {"host": host, "port": int(port), "offset": int(offset)}
            )

        return res

    def _parse_slave(response):
        host, port, status, offset = response[1:]

        return {
            "role": role,
            "host": host,
            "port": port,
            "status": status,
            "offset": offset,
        }

    def _parse_sentinel(response):
        return {"role": role, "masters": response[1:]}

    parser = {
        "master": _parse_master,
        "slave": _parse_slave,
        "sentinel": _parse_sentinel,
    }[role]

    return parser(response)


class ServerCommandMixin:
    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict("BGREWRITEAOF BGSAVE", lambda r: True),
        string_keys_to_dict("FLUSHALL FLUSHDB SAVE " "SHUTDOWN SLAVEOF", bool_ok),
        {
            "ROLE": parse_role,
            "SLOWLOG GET": parse_slowlog_get,
            "SLOWLOG LEN": int,
            "SLOWLOG RESET": bool_ok,
            "CLIENT GETNAME": lambda r: r and nativestr(r),
            "CLIENT KILL": bool_ok,
            "CLIENT LIST": parse_client_list,
            "CLIENT SETNAME": bool_ok,
            "CLIENT PAUSE": bool_ok,
            "CONFIG GET": parse_config_get,
            "CONFIG RESETSTAT": bool_ok,
            "CONFIG SET": bool_ok,
            "DEBUG OBJECT": parse_debug_object,
            "INFO": parse_info,
            "LASTSAVE": timestamp_to_datetime,
            "TIME": lambda x: (int(x[0]), int(x[1])),
        },
    )

    async def bgrewriteaof(self):
        """Tell the Redis server to rewrite the AOF file from data in memory"""

        return await self.execute_command("BGREWRITEAOF")

    async def bgsave(self):
        """
        Tells the Redis server to save its data to disk.  Unlike save(),
        this method is asynchronous and returns immediately.
        """

        return await self.execute_command("BGSAVE")

    async def client_kill(self, address):
        """Disconnects the client at ``address`` (ip:port)"""

        return await self.execute_command("CLIENT KILL", address)

    async def client_list(self):
        """Returns a list of currently connected clients"""

        return await self.execute_command("CLIENT LIST")

    async def client_getname(self):
        """Returns the current connection name"""

        return await self.execute_command("CLIENT GETNAME")

    async def client_setname(self, name):
        """Sets the current connection name"""

        return await self.execute_command("CLIENT SETNAME", name)

    async def client_pause(self, timeout=0):
        """
        Suspends all the Redis clients for the specified amount of time
        (in milliseconds).
        """

        return await self.execute_command("CLIENT PAUSE", timeout)

    async def config_get(self, pattern="*"):
        """Returns a dictionary of configuration based on the ``pattern``"""

        return await self.execute_command("CONFIG GET", pattern)

    async def config_set(self, name, value):
        """Sets config item ``name`` to ``value``"""

        return await self.execute_command("CONFIG SET", name, value)

    async def config_resetstat(self):
        """Resets runtime statistics"""

        return await self.execute_command("CONFIG RESETSTAT")

    async def config_rewrite(self):
        """
        Rewrites config file with the minimal change to reflect running config
        """

        return await self.execute_command("CONFIG REWRITE")

    async def dbsize(self):
        """Returns the number of keys in the current database"""

        return await self.execute_command("DBSIZE")

    async def debug_object(self, key):
        """Returns version specific meta information about a given key"""

        return await self.execute_command("DEBUG OBJECT", key)

    async def flushall(self):
        """Deletes all keys in all databases on the current host"""

        return await self.execute_command("FLUSHALL")

    async def flushdb(self):
        """Deletes all keys in the current database"""

        return await self.execute_command("FLUSHDB")

    async def info(self, section=None):
        """
        Returns a dictionary containing information about the Redis server

        The ``section`` option can be used to select a specific section
        of information

        The section option is not supported by older versions of Redis Server,
        and will generate ResponseError
        """

        if section is None:
            return await self.execute_command("INFO")
        else:
            return await self.execute_command("INFO", section)

    async def lastsave(self):
        """
        Returns a Python datetime object representing the last time the
        Redis database was saved to disk
        """

        return await self.execute_command("LASTSAVE")

    async def save(self):
        """
        Tells the Redis server to save its data to disk,
        blocking until the save is complete
        """

        return await self.execute_command("SAVE")

    async def shutdown(self):
        """Stops Redis server"""
        try:
            await self.execute_command("SHUTDOWN")
        except ConnectionError:
            # a ConnectionError here is expected

            return
        raise RedisError("SHUTDOWN seems to have failed.")

    async def slaveof(self, host=None, port=None):
        """
        Sets the server to be a replicated slave of the instance identified
        by the ``host`` and ``port``. If called without arguments, the
        instance is promoted to a master instead.
        """

        if host is None and port is None:
            return await self.execute_command("SLAVEOF", b("NO"), b("ONE"))

        return await self.execute_command("SLAVEOF", host, port)

    async def slowlog_get(self, num=None):
        """
        Gets the entries from the slowlog. If ``num`` is specified, get the
        most recent ``num`` items.
        """
        args = ["SLOWLOG GET"]

        if num is not None:
            args.append(num)

        return await self.execute_command(*args)

    async def slowlog_len(self):
        """Gets the number of items in the slowlog"""

        return await self.execute_command("SLOWLOG LEN")

    async def slowlog_reset(self):
        """Removes all items in the slowlog"""

        return await self.execute_command("SLOWLOG RESET")

    async def time(self):
        """
        Returns the server time as a 2-item tuple of ints:
        (seconds since epoch, microseconds into this second).
        """

        return await self.execute_command("TIME")

    async def role(self):
        """
        Provides information on the role of a Redis instance in the context of replication,
        by returning if the instance is currently a master, slave, or sentinel.
        The command also returns additional information about the state of the replication
        (if the role is master or slave)
        or the list of monitored master names (if the role is sentinel).
        :return:
        """

        return await self.execute_command("ROLE")

    async def lolwut(self, *version_numbers, **kwargs):
        """
        Get the Redis version and a piece of generative computer art

        .. versionadded:: 2.1.0
        """

        if version_numbers:
            return await self.execute_command(
                "LOLWUT VERSION", *version_numbers, **kwargs
            )
        else:
            return await self.execute_command("LOLWUT", **kwargs)


class ClusterServerCommandMixin(ServerCommandMixin):
    NODES_FLAGS = dict_merge(
        list_keys_to_dict(["SHUTDOWN", "SLAVEOF", "CLIENT SETNAME"], NodeFlag.BLOCKED),
        list_keys_to_dict(["FLUSHALL", "FLUSHDB"], NodeFlag.ALL_MASTERS),
        list_keys_to_dict(
            [
                "SLOWLOG LEN",
                "SLOWLOG RESET",
                "SLOWLOG GET",
                "TIME",
                "SAVE",
                "LASTSAVE",
                "DBSIZE",
                "CONFIG RESETSTAT",
                "CONFIG REWRITE",
                "CONFIG GET",
                "CONFIG SET",
                "CLIENT KILL",
                "CLIENT LIST",
                "CLIENT GETNAME",
                "INFO",
                "BGSAVE",
                "BGREWRITEAOF",
            ],
            NodeFlag.ALL_NODES,
        ),
    )

    RESULT_CALLBACKS = dict_merge(
        list_keys_to_dict(
            [
                "CONFIG GET",
                "CONFIG SET",
                "SLOWLOG GET",
                "CLIENT KILL",
                "INFO",
                "BGREWRITEAOF",
                "BGSAVE",
                "CLIENT LIST",
                "CLIENT GETNAME",
                "CONFIG RESETSTAT",
                "CONFIG REWRITE",
                "DBSIZE",
                "LASTSAVE",
                "SAVE",
                "SLOWLOG LEN",
                "SLOWLOG RESET",
                "TIME",
                "FLUSHALL",
                "FLUSHDB",
            ],
            lambda res: res,
        )
    )
